@@ -1,5 +1,5 @@
 /* decl.cc -- Lower D frontend declarations to GCC trees.
-   Copyright (C) 2006-2023 Free Software Foundation, Inc.
+   Copyright (C) 2006-2025 Free Software Foundation, Inc.
 
 GCC is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
@@ -36,6 +36,7 @@ along with GCC; see the file COPYING3.  If not see
 #include "dmd/nspace.h"
 #include "dmd/target.h"
 #include "dmd/template.h"
+#include "dmd/typinf.h"
 
 #include "tree.h"
 #include "tree-iterator.h"
@@ -69,11 +70,11 @@ const char *
 d_mangle_decl (Dsymbol *decl)
 {
   if (decl->isFuncDeclaration ())
-    return mangleExact ((FuncDeclaration *) decl);
+    return dmd::mangleExact ((FuncDeclaration *) decl);
   else
     {
       OutBuffer buf;
-      mangleToBuffer (decl, &buf);
+      dmd::mangleToBuffer (decl, buf);
       return buf.extractChars ();
     }
 }
@@ -255,18 +256,18 @@ public:
 
   void visit (Module *d) final override
   {
-    if (d->semanticRun >= PASS::obj)
+    if (d->semanticRun () >= PASS::obj)
       return;
 
     build_module_tree (d);
-    d->semanticRun = PASS::obj;
+    d->semanticRun (PASS::obj);
   }
 
   /* Write the imported symbol to debug.  */
 
   void visit (Import *d) final override
   {
-    if (d->semanticRun >= PASS::obj)
+    if (d->semanticRun () >= PASS::obj)
       return;
 
     /* Implements import declarations by telling the debug back-end we are
@@ -319,7 +320,15 @@ public:
 					      false, false);
       }
 
-    d->semanticRun = PASS::obj;
+    d->semanticRun (PASS::obj);
+  }
+
+  /* Finish a top-level `asm` definition.  */
+
+  void visit (CAsmDeclaration *d) final override
+  {
+    tree asm_str = build_expr (d->code);
+    symtab->finalize_toplevel_asm (asm_str);
   }
 
   /* Expand any local variables found in tuples.  */
@@ -342,7 +351,7 @@ public:
 
   void visit (AttribDeclaration *d) final override
   {
-    Dsymbols *ds = d->include (NULL);
+    Dsymbols *ds = dmd::include (d, NULL);
 
     if (!ds)
       return;
@@ -392,7 +401,7 @@ public:
 
   void visit (Nspace *d) final override
   {
-    if (isError (d) || !d->members)
+    if (dmd::isError (d) || !d->members)
       return;
 
     for (size_t i = 0; i < d->members->length; i++)
@@ -437,7 +446,7 @@ public:
 
   void visit (TemplateInstance *d) final override
   {
-    if (isError (d)|| !d->members)
+    if (dmd::isError (d)|| !d->members)
       return;
 
     if (!d->needsCodegen ())
@@ -451,7 +460,7 @@ public:
 
   void visit (TemplateMixin *d) final override
   {
-    if (isError (d)|| !d->members)
+    if (dmd::isError (d)|| !d->members)
       return;
 
     for (size_t i = 0; i < d->members->length; i++)
@@ -463,7 +472,7 @@ public:
 
   void visit (StructDeclaration *d) final override
   {
-    if (d->semanticRun >= PASS::obj)
+    if (d->semanticRun () >= PASS::obj)
       return;
 
     if (d->type->ty == TY::Terror)
@@ -511,7 +520,7 @@ public:
     if (d->xhash)
       this->build_dsymbol (d->xhash);
 
-    d->semanticRun = PASS::obj;
+    d->semanticRun (PASS::obj);
   }
 
   /* Finish semantic analysis of functions in vtbl for class CD.  */
@@ -529,11 +538,11 @@ public:
 	  continue;
 
 	/* Ensure function has a return value.  */
-	if (!fd->functionSemantic ())
+	if (!dmd::functionSemantic (fd))
 	  has_errors = true;
 
 	/* No name hiding to check for.  */
-	if (!d->isFuncHidden (fd) || fd->isFuture ())
+	if (!dmd::isFuncHidden (d, fd) || fd->isFuture ())
 	  continue;
 
 	/* The function fd is hidden from the view of the class.
@@ -553,20 +562,21 @@ public:
 	    if (fd2->isFuture ())
 	      continue;
 
-	    if (fd->leastAsSpecialized (fd2, NULL) != MATCH::nomatch
-		|| fd2->leastAsSpecialized (fd, NULL) != MATCH::nomatch)
-	      {
-		error_at (make_location_t (fd->loc), "use of %qs",
-			  fd->toPrettyChars ());
-		inform (make_location_t (fd2->loc), "is hidden by %qs",
-			fd2->toPrettyChars ());
-		inform (make_location_t (d->loc),
-			"use %<alias %s = %s.%s;%> to introduce base class "
-			"overload set", fd->toChars (),
-			fd->parent->toChars (), fd->toChars ());
-		has_errors = true;
-		break;
-	      }
+	    if (dmd::leastAsSpecialized (fd, fd2, NULL) == MATCH::nomatch
+		&& dmd::leastAsSpecialized (fd2, fd, NULL) == MATCH::nomatch)
+	      continue;
+
+	    /* Hiding detected; same name, overlapping specializations.  */
+	    error_at (make_location_t (fd->loc), "use of %qs",
+		      fd->toPrettyChars ());
+	    inform (make_location_t (fd2->loc), "is hidden by %qs",
+		    fd2->toPrettyChars ());
+	    inform (make_location_t (d->loc),
+		    "use %<alias %s = %s.%s;%> to introduce base class "
+		    "overload set", fd->toChars (),
+		    fd->parent->toChars (), fd->toChars ());
+	    has_errors = true;
+	    break;
 	  }
       }
 
@@ -578,7 +588,7 @@ public:
 
   void visit (ClassDeclaration *d) final override
   {
-    if (d->semanticRun >= PASS::obj)
+    if (d->semanticRun () >= PASS::obj)
       return;
 
     if (d->type->ty == TY::Terror)
@@ -607,7 +617,7 @@ public:
 
     /* Generate C symbols.  */
     d->csym = get_classinfo_decl (d);
-    Dsymbol *vtblsym = d->vtblSymbol ();
+    Dsymbol *vtblsym = dmd::vtblSymbol (d);
     vtblsym->csym = get_vtable_decl (d);
     tree sinit = aggregate_initializer_decl (d);
 
@@ -644,7 +654,7 @@ public:
       = build_constructor (TREE_TYPE (vtblsym->csym), elms);
     d_finish_decl (vtblsym->csym);
 
-    d->semanticRun = PASS::obj;
+    d->semanticRun (PASS::obj);
   }
 
   /* Write out compiler generated TypeInfo and vtables for the given interface
@@ -652,7 +662,7 @@ public:
 
   void visit (InterfaceDeclaration *d) final override
   {
-    if (d->semanticRun >= PASS::obj)
+    if (d->semanticRun () >= PASS::obj)
       return;
 
     if (d->type->ty == TY::Terror)
@@ -687,7 +697,7 @@ public:
     DECL_INITIAL (d->csym) = layout_classinfo (d);
     d_finish_decl (d->csym);
 
-    d->semanticRun = PASS::obj;
+    d->semanticRun (PASS::obj);
   }
 
   /* Write out compiler generated TypeInfo and initializer for the given
@@ -695,10 +705,10 @@ public:
 
   void visit (EnumDeclaration *d) final override
   {
-    if (d->semanticRun >= PASS::obj)
+    if (d->semanticRun () >= PASS::obj)
       return;
 
-    if (d->errors || d->type->ty == TY::Terror)
+    if (d->errors () || d->type->ty == TY::Terror)
       {
 	error_at (make_location_t (d->loc),
 		  "had semantic errors when compiling");
@@ -718,7 +728,7 @@ public:
       create_typeinfo (d->type, NULL);
 
     TypeEnum *tc = d->type->isTypeEnum ();
-    if (tc->sym->members && !d->type->isZeroInit ())
+    if (tc->sym->members && !dmd::isZeroInit (d->type))
       {
 	/* Generate static initializer.  */
 	d->sinit = enum_initializer_decl (d);
@@ -726,7 +736,7 @@ public:
 	d_finish_decl (d->sinit);
       }
 
-    d->semanticRun = PASS::obj;
+    d->semanticRun (PASS::obj);
   }
 
   /* Finish up a variable declaration and push it into the current scope.
@@ -734,7 +744,7 @@ public:
 
   void visit (VarDeclaration *d) final override
   {
-    if (d->semanticRun >= PASS::obj)
+    if (d->semanticRun () >= PASS::obj)
       return;
 
     if (d->type->ty == TY::Terror)
@@ -752,7 +762,7 @@ public:
 	    && d->_init && !d->_init->isVoidInitializer ())
 	  {
 	    /* Evaluate RHS for side effects first.  */
-	    Expression *ie = initializerToExpression (d->_init);
+	    Expression *ie = dmd::initializerToExpression (d->_init);
 	    add_stmt (build_expr (ie));
 
 	    Expression *e = d->type->defaultInitLiteral (d->loc);
@@ -772,7 +782,7 @@ public:
       {
 	/* Do not store variables we cannot take the address of,
 	   but keep the values for purposes of debugging.  */
-	if (d->type->isscalar () && !d->type->hasPointers ())
+	if (d->type->isScalar () && !dmd::hasPointers (d->type))
 	  {
 	    tree decl = get_symbol_decl (d);
 	    d_pushdecl (decl);
@@ -794,7 +804,7 @@ public:
 	  return;
 
 	/* How big a symbol can be should depend on back-end.  */
-	tree size = build_integer_cst (d->type->size (d->loc),
+	tree size = build_integer_cst (dmd::size (d->type, d->loc),
 				       build_ctype (Type::tsize_t));
 	if (!valid_constant_size_p (size))
 	  {
@@ -807,11 +817,12 @@ public:
 	    /* Use the explicit initializer, this includes `void`.  */
 	    if (!d->_init->isVoidInitializer ())
 	      {
-		Expression *e = initializerToExpression (d->_init, d->type);
+		Expression *e =
+		  dmd::initializerToExpression (d->_init, d->type);
 		DECL_INITIAL (decl) = build_expr (e, true);
 	      }
 	  }
-	else if (!d->type->isZeroInit ())
+	else if (!dmd::isZeroInit (d->type))
 	  {
 	    /* Use default initializer for the type.  */
 	    if (TypeStruct *ts = d->type->isTypeStruct ())
@@ -844,7 +855,7 @@ public:
 	    tree decl = get_symbol_decl (d);
 
 	    ExpInitializer *vinit = d->_init->isExpInitializer ();
-	    Expression *ie = initializerToExpression (vinit);
+	    Expression *ie = dmd::initializerToExpression (vinit);
 	    tree exp = build_expr (ie);
 
 	    /* Maybe put variable on list of things needing destruction.  */
@@ -878,7 +889,7 @@ public:
 	  }
       }
 
-    d->semanticRun = PASS::obj;
+    d->semanticRun (PASS::obj);
   }
 
   /* Generate and compile a static TypeInfo declaration, but only if it is
@@ -886,16 +897,16 @@ public:
 
   void visit (TypeInfoDeclaration *d) final override
   {
-    if (d->semanticRun >= PASS::obj)
+    if (d->semanticRun () >= PASS::obj)
       return;
 
-    if (speculative_type_p (d->tinfo))
+    if (dmd::isSpeculativeType (d->tinfo))
       return;
 
     tree t = get_typeinfo_decl (d);
     DECL_INITIAL (t) = layout_typeinfo (d);
     d_finish_decl (t);
-    d->semanticRun = PASS::obj;
+    d->semanticRun (PASS::obj);
   }
 
   /* Finish up a function declaration and compile it all the way
@@ -904,7 +915,7 @@ public:
   void visit (FuncDeclaration *d) final override
   {
     /* Already generated the function.  */
-    if (d->semanticRun >= PASS::obj)
+    if (d->semanticRun () >= PASS::obj)
       return;
 
     /* Don't emit any symbols from gcc.attributes module.  */
@@ -946,12 +957,12 @@ public:
       }
 
     /* Ensure all semantic passes have run.  */
-    if (d->semanticRun < PASS::semantic3)
+    if (d->semanticRun () < PASS::semantic3)
       {
 	gcc_assert (!doing_semantic_analysis_p);
 
 	doing_semantic_analysis_p = true;
-	d->functionSemantic3 ();
+	dmd::functionSemantic3 (d);
 	Module::runDeferredSemantic3 ();
 	doing_semantic_analysis_p = false;
       }
@@ -960,8 +971,8 @@ public:
       return;
 
     /* Start generating code for this function.  */
-    gcc_assert (d->semanticRun == PASS::semantic3done);
-    d->semanticRun = PASS::obj;
+    gcc_assert (d->semanticRun () == PASS::semantic3done);
+    d->semanticRun (PASS::obj);
 
     /* Duplicated FuncDeclarations map to the same symbol.  Check if this
        is the one declaration which will be emitted.  */
@@ -976,7 +987,7 @@ public:
 	return;
       }
 
-    if (global.params.verbose)
+    if (global.params.v.verbose)
       message ("function  %s", d->toPrettyChars ());
 
     tree old_context = start_function (d);
@@ -1101,10 +1112,10 @@ build_decl_tree (Dsymbol *d)
   location_t saved_location = input_location;
 
   /* Set input location, empty DECL_SOURCE_FILE can crash debug generator.  */
-  if (d->loc.filename)
+  if (d->loc.filename ())
     input_location = make_location_t (d->loc);
   else
-    input_location = make_location_t (Loc ("<no_file>", 1, 0));
+    input_location = make_location_t (Loc::singleFilename ("<no_file>"));
 
   DeclVisitor v = DeclVisitor ();
   v.build_dsymbol (d);
@@ -1168,7 +1179,7 @@ maybe_build_decl_tree (Declaration *decl)
 
   /* Still running semantic analysis on declaration, or it has already had its
      code generated.  */
-  if (doing_semantic_analysis_p || decl->semanticRun >= PASS::obj)
+  if (doing_semantic_analysis_p || decl->semanticRun () >= PASS::obj)
     return decl->csym;
 
   if (error_operand_p (decl->csym))
@@ -1235,10 +1246,10 @@ get_symbol_decl (Declaration *decl)
       /* CONST_DECL was initially intended for enumerals and may be used for
 	 scalars in general, but not for aggregates.  Here a non-constant
 	 value is generated anyway so as its value can be used.  */
-      if (!vd->canTakeAddressOf () && !vd->type->isscalar ())
+      if (!vd->canTakeAddressOf () && !vd->type->isScalar ())
 	{
 	  gcc_assert (vd->_init && !vd->_init->isVoidInitializer ());
-	  Expression *ie = initializerToExpression (vd->_init);
+	  Expression *ie = dmd::initializerToExpression (vd->_init);
 	  decl->csym = build_expr (ie, false);
 	  return decl->csym;
 	}
@@ -1249,7 +1260,7 @@ get_symbol_decl (Declaration *decl)
   if (fd)
     {
       /* Run full semantic on functions we need to know about.  */
-      if (!fd->functionSemantic ())
+      if (!dmd::functionSemantic (fd))
 	{
 	  decl->csym = error_mark_node;
 	  return decl->csym;
@@ -1296,9 +1307,9 @@ get_symbol_decl (Declaration *decl)
 	  /* Cannot make an expression out of a void initializer.  */
 	  gcc_assert (vd->_init && !vd->_init->isVoidInitializer ());
 	  /* Non-scalar manifest constants have already been dealt with.  */
-	  gcc_assert (vd->type->isscalar ());
+	  gcc_assert (vd->type->isScalar ());
 
-	  Expression *ie = initializerToExpression (vd->_init);
+	  Expression *ie = dmd::initializerToExpression (vd->_init);
 	  DECL_INITIAL (decl->csym) = build_expr (ie, true);
 	}
 
@@ -1313,7 +1324,7 @@ get_symbol_decl (Declaration *decl)
       /* `const` applies to data that cannot be changed by the const reference
 	 to that data. It may, however, be changed by another reference to that
 	 same data.  */
-      if (vd->isConst () && !vd->isDataseg ())
+      if (vd->isConst () && !vd->isResult () && !vd->isDataseg ())
 	TREE_READONLY (decl->csym) = 1;
     }
 
@@ -1572,7 +1583,7 @@ get_symbol_decl (Declaration *decl)
   /* Symbol is going in thread local storage.  */
   if (decl->isThreadlocal () && !DECL_ARTIFICIAL (decl->csym))
     {
-      if (global.params.vtls)
+      if (global.params.v.tls)
 	message (decl->loc, "`%s` is thread local", decl->toChars ());
 
       set_decl_tls_model (decl->csym, decl_default_tls_model (decl->csym));
@@ -2082,7 +2093,7 @@ start_function (FuncDeclaration *fd)
   allocate_struct_function (fndecl, false);
 
   /* Store the end of the function.  */
-  if (fd->endloc.filename)
+  if (fd->endloc.filename ())
     cfun->function_end_locus = make_location_t (fd->endloc);
   else
     cfun->function_end_locus = DECL_SOURCE_LOCATION (fndecl);
@@ -2199,9 +2210,9 @@ get_vtable_decl (ClassDeclaration *decl)
   tree ident = mangle_internal_decl (decl, "__vtbl", "Z");
   /* Note: Using a static array type for the VAR_DECL, the DECL_INITIAL value
      will have a different type.  However the back-end seems to accept this.  */
-  tree type = build_ctype (Type::tvoidptr->sarrayOf (decl->vtbl.length));
+  tree type = build_ctype (dmd::sarrayOf (Type::tvoidptr, decl->vtbl.length));
 
-  Dsymbol *vtblsym = decl->vtblSymbol ();
+  Dsymbol *vtblsym = dmd::vtblSymbol (decl);
   vtblsym->csym = declare_extern_var (ident, type);
   DECL_LANG_SPECIFIC (vtblsym->csym) = build_lang_decl (NULL);
 
@@ -2382,6 +2393,12 @@ aggregate_initializer_decl (AggregateDeclaration *decl)
       SET_DECL_ALIGN (sinit, sd->alignment.get () * BITS_PER_UNIT);
       DECL_USER_ALIGN (sinit) = true;
     }
+  else if (sd == NULL)
+    {
+      /* Alignment of class is determined its biggest field alignment.  */
+      SET_DECL_ALIGN (sinit, decl->alignsize * BITS_PER_UNIT);
+      DECL_USER_ALIGN (sinit) = true;
+    }
 
   decl->sinit = sinit;
   return sinit;
@@ -2392,10 +2409,10 @@ aggregate_initializer_decl (AggregateDeclaration *decl)
 tree
 layout_class_initializer (ClassDeclaration *cd)
 {
-  NewExp *ne = NewExp::create (cd->loc, NULL, cd->type, NULL);
+  NewExp *ne = NewExp::create (cd->loc, NULL, NULL, cd->type, NULL);
   ne->type = cd->type;
 
-  Expression *e = ne->ctfeInterpret ();
+  Expression *e = dmd::ctfeInterpret (ne);
   gcc_assert (e->op == EXP::classReference);
 
   return build_class_instance (e->isClassReferenceExp ());
@@ -2406,7 +2423,7 @@ layout_struct_initializer (StructDeclaration *sd)
 {
   StructLiteralExp *sle = StructLiteralExp::create (sd->loc, sd, NULL);
 
-  if (!sd->fill (sd->loc, *sle->elements, true))
+  if (!dmd::fill (sd, sd->loc, *sle->elements, true))
     gcc_unreachable ();
 
   sle->type = sd->type;
