@@ -1,5 +1,5 @@
 /* modules.cc -- D module initialization and termination.
-   Copyright (C) 2013-2024 Free Software Foundation, Inc.
+   Copyright (C) 2013-2025 Free Software Foundation, Inc.
 
 GCC is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
@@ -30,7 +30,6 @@ along with GCC; see the file COPYING3.  If not see
 #include "function.h"
 #include "cgraph.h"
 #include "stor-layout.h"
-#include "debug.h"
 #include "toplev.h"
 #include "target.h"
 #include "common/common-target.h"
@@ -144,13 +143,13 @@ get_internal_fn (tree ident, const Visibility &visibility)
       name = IDENTIFIER_POINTER (s);
     }
 
-  FuncDeclaration *fd = FuncDeclaration::genCfunc (NULL, Type::tvoid,
-						   Identifier::idPool (name));
+  FuncDeclaration *fd = dmd::genCfunc (NULL, Type::tvoid,
+				       Identifier::idPool (name));
   fd->isGenerated (true);
-  fd->loc = Loc (mod->srcfile.toChars (), 1, 0);
+  fd->loc = Loc::singleFilename (mod->srcfile.toChars ());
   fd->parent = mod;
   fd->visibility = visibility;
-  fd->semanticRun = PASS::semantic3done;
+  fd->semanticRun (PASS::semantic3done);
 
   return fd;
 }
@@ -928,14 +927,6 @@ d_finish_compilation (tree *vec, int len)
   /* Complete all generated thunks.  */
   symtab->process_same_body_aliases ();
 
-  /* Output debug information for all type declarations in this unit.  */
-  for (int i = 0; i < len; i++)
-    {
-      tree decl = vec[i];
-      if (TREE_CODE (decl) == TYPE_DECL)
-	debug_hooks->type_decl (decl, false);
-    }
-
   /* Process all file scopes in this compilation, and the external_scope,
      through wrapup_global_declarations.  */
   for (int i = 0; i < len; i++)
